import { BaseEvent } from '#application/base/baseEvent'
import { UUID } from '#application/types'
import SpriteRepository from '#repositories/spriteRepository'
import { SpriteAction } from '#entities/spriteAction'
import sharp from 'sharp'
import fs from 'fs'

interface SpriteImage {
  url: string
  offset: {
    x: number
    y: number
  }
}

type Payload = {
  id: UUID
  name: string
  spriteActions: Array<{
    action: string
    sprites: SpriteImage[]
    originX: number
    originY: number
    frameRate: number
  }>
}

export default class SpriteUpdateEvent extends BaseEvent {
  public listen(): void {
    this.socket.on('gm:sprite:update', this.handleEvent.bind(this))
  }

  private async handleEvent(data: Payload, callback: (success: boolean) => void): Promise<void> {
    try {
      if (!(await this.isCharacterGM())) return

      const spriteRepository = new SpriteRepository()
      const sprite = await spriteRepository.getById(data.id)
      if (!sprite) return callback(false)

      await spriteRepository.getEntityManager().populate(sprite, ['spriteActions'])

      // Update sprite in database
      await sprite.setName(data.name).save()

      // First verify all sprite sheets can be generated
      for (const actionData of data.spriteActions) {
        if (!await this.generateSpriteSheet(actionData.sprites, sprite.getId(), actionData.action)) {
          return callback(false)
        }
      }

      const existingActions = sprite.getSpriteActions()

      // Remove existing actions only after confirming sprite sheets generated successfully
      for (const existingAction of existingActions) {
        await spriteRepository.getEntityManager().removeAndFlush(existingAction)
      }

      // Create new actions
      for (const actionData of data.spriteActions) {
        const spriteAction = new SpriteAction()
        spriteAction.setSprite(sprite)
        sprite.getSpriteActions().add(spriteAction)

        spriteAction
          .setAction(actionData.action)
          .setSprites(actionData.sprites)
          .setOriginX(actionData.originX)
          .setOriginY(actionData.originY)
          .setFrameWidth(await this.calculateMaxWidth(actionData.sprites))
          .setFrameHeight(await this.calculateMaxHeight(actionData.sprites))
          .setFrameRate(actionData.frameRate)

        await spriteRepository.getEntityManager().persistAndFlush(spriteAction)
      }

      return callback(true)
    } catch (error) {
      console.error(`Error updating sprite ${data.id}:`, error)
      return callback(false)
    }
  }

  private async calculateMaxWidth(sprites: SpriteImage[]): Promise<number> {
    if (!sprites.length) return 0

    const widths = await Promise.all(
      sprites.map(async (base64) => {
        const uri = base64.url.split(';base64,').pop()
        if (!uri) return 0

        const imgBuffer = Buffer.from(uri, 'base64')
        const image = await sharp(imgBuffer).metadata()
        return image.width ?? 0
      })
    )

    return Math.max(...widths)
  }

  private async calculateMaxHeight(sprites: SpriteImage[]): Promise<number> {
    if (!sprites.length) return 0

    const heights = await Promise.all(
      sprites.map(async (base64) => {
        const uri = base64.url.split(';base64,').pop()
        if (!uri) return 0

        const imgBuffer = Buffer.from(uri, 'base64')
        const image = await sharp(imgBuffer).metadata()
        return image.height ?? 0
      })
    )

    return Math.max(...heights)
  }

  private async generateSpriteSheet(sprites: SpriteImage[], spriteId: string, action: string): Promise<boolean> {
    try {
      // Skip if no sprites
      if (!sprites.length) return true

      // Process all base64 images to buffers and get their dimensions
      const imageData = await Promise.all(
        sprites.map(async (sprite) => {
          const uri = sprite.url.split(';base64,').pop()
          if (!uri) throw new Error('Invalid base64 image')
          const buffer = Buffer.from(uri, 'base64')
          const metadata = await sharp(buffer).metadata()
          return { 
            buffer, 
            width: metadata.width ?? 0, 
            height: metadata.height ?? 0,
            offsetX: sprite.offset?.x ?? 0,
            offsetY: sprite.offset?.y ?? 0
          }
        })
      )

      // Find the largest dimensions including offsets
      const effectiveDimensions = imageData.map(({ width, height, offsetX, offsetY }) => ({
        width: width + Math.abs(offsetX),
        height: height + Math.abs(offsetY),
        top: offsetY >= 0 ? offsetY : 0,
        bottom: offsetY < 0 ? Math.abs(offsetY) : 0
      }))
      const maxWidth = Math.max(...effectiveDimensions.map(d => d.width))
      const maxHeight = Math.max(...effectiveDimensions.map(d => d.height))
      const maxTop = Math.max(...effectiveDimensions.map(d => d.top))
      const maxBottom = Math.max(...effectiveDimensions.map(d => d.bottom))

      // Calculate total height needed to accommodate all sprites with their offsets
      const totalHeight = maxHeight + maxTop + maxBottom

      // Extend all images to match the largest dimensions without resizing
      const resizedBuffers = await Promise.all(
        imageData.map(async ({ buffer, width, height, offsetX, offsetY }) => {
          // Calculate the position of the image within the frame
          const left = offsetX >= 0 ? offsetX : 0
          // Position from bottom edge while respecting offsetY
          const verticalOffset = totalHeight - height - (offsetY >= 0 ? offsetY : 0)

          // Create a blank canvas of the maximum size
          return await sharp({
            create: {
              width: maxWidth,
              height: totalHeight,
              channels: 4,
              background: { r: 0, g: 0, b: 0, alpha: 0 }
            }
          })
          .composite([
            {
              input: buffer,
              left: left,
              top: verticalOffset
            }
          ])
          .png()
          .toBuffer()
        })
      )

      // Create sprite sheet by combining images horizontally
      const spriteSheet = await sharp({
        create: {
          width: maxWidth * sprites.length,
          height: totalHeight,
          channels: 4,
          background: { r: 0, g: 0, b: 0, alpha: 0 }
        }
      })
        .composite(
          resizedBuffers.map((buffer, index) => ({
            input: buffer,
            left: index * maxWidth,
            top: 0
          }))
        )
        .png()
        .toBuffer()

      // Ensure directory exists
      const dir = `public/sprites/${spriteId}`
      await fs.promises.mkdir(dir, { recursive: true })

      // Save the sprite sheet
      await fs.promises.writeFile(`${dir}/${action}.png`, spriteSheet)

      return true
    } catch (error) {
      console.error('Error generating sprite sheet:', error)
      return false
    }
  }
}
